#!/usr/bin/python

# Copyright (C) 2010 Poltorak Serguei
#
# Z-Stick Updater is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Z-Stick Updater is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Z-Stick Updater.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import optparse
import logging

import SerialAPI
import ZLogging
import IntelHex

class Programmer:
	### Timing configuration in seconds
	#
	# queue loop wait time
	dt = 0.1

	SOF = 0x01
	ACK = 0x06
	NAK = 0x10
	CRC = 0x30
	TMT = 0x70
	
	# Zensys Serial API protocol headers
	zSOF = 0x01
	zACK = 0x06
	zNAK = 0x15
	zCAN = 0x18
	
	TIMEOUT = 0.5
	TRY_COUNT_MAX = 3
	
	REBOOT_WAIT = 3
	TIMEOUT_WRONG_PAGE_WAIT = 1

	def PurgeBuffer(self):
		while self.serialPort.inWaiting() > 0:
			self.serialPort.Read(1)
			time.sleep(self.dt)

	def Wait(self, period):
		logging.serial.info("Waiting for relaxation")
		time.sleep(period)
	
	def __init__(self, serialPort, file, freq):
		self.file = file
		self.freq = freq

		# opening serial port
		self.serialPort = serialPort
		self.serialPort.Exit = self.Exit
		self.serialPort.Open()

		# Purge buffer
		self.PurgeBuffer()

		wpages = IntelHex.ReadHex(self.file)
		
		# Change frequency if needed
		if self.freq == 'EU':
			wpages[0x7f][0x89] = 0x0a
			wpages[0x7f][0x8b] = 0x08
			wpages[0x7f][0xb1] = 0x00
		elif self.freq == 'RU':
			wpages[0x7f][0x89] = 0x07
			wpages[0x7f][0x8b] = 0x05
			wpages[0x7f][0xb1] = 0x0a
		elif self.freq == 'US':
			wpages[0x7f][0x89] = 0x02
			wpages[0x7f][0x8b] = 0x00
			wpages[0x7f][0xb1] = 0x01

		# try to put stick in reflash mode
		packet = [self.zSOF, 0x03, 0x00, 0xf1, 0x0d] # Zensys Serial API protocol [SOF, len, REQ, ZMEReflashFunc, CRC]

		tryCount = 1
		while tryCount:
			self.PurgeBuffer()
			
			self.serialPort.Write(packet) # send packet
			logging.serial.info("Putting stick into reflash mode")
			
			timeout = 0
			while 1:
				time.sleep(self.dt)
				timeout += self.dt
				if self.serialPort.inWaiting() > 0:
					resp = self.serialPort.Read(1)
					if resp[0] == self.NAK or resp[0] == self.CRC or resp[0] == self.TMT:
						logging.serial.info("Z-Stick is probably alredy in reflash mode")
						tryCount = 0
						self.Wait(self.TIMEOUT_WRONG_PAGE_WAIT)
						break;
					elif resp[0] == self.ACK:
						logging.serial.info("Got Ack. Z-Stick is now in reflash mode (if supported)")
						tryCount = 0
						self.Wait(self.REBOOT_WAIT)
						break # go further
					elif resp[0] == self.zNAK or resp[0] == self.zCAN:
						logging.serial.error("Z-Stick is not in reflash mode, but is busy")
						if tryCount >= self.TRY_COUNT_MAX:
							exit(7)
						# try again
					else:
						logging.serial.error("Unknown response %i" % (resp[0]))
						if tryCount >= self.TRY_COUNT_MAX:
							exit(9)
						# try again
					tryCount += 1
					break
				if timeout >= self.TIMEOUT:
					logging.serial.error("Operation timeout")
					if tryCount >= self.TRY_COUNT_MAX:
						logging.serial.info("Z-Stick is probably alredy in reflash mode")
						tryCount = 0
						self.Wait(self.TIMEOUT_WRONG_PAGE_WAIT)
						break # go further
					tryCount += 1
					break

		self.PurgeBuffer()
		
		# write pages
		for p in range(len(wpages)-1, -1, -1):
			buf = [self.SOF, p]
			buf.extend(wpages[p])
			buf.append(sum(buf[1:]) & 0xff) # page checksum

			tryCount = 1
			while tryCount:
				self.serialPort.Write(buf) # send
				logging.serial.debug("Writing page %i: %s" % (p, buf))
			
				timeout = 0
				while 1:
					time.sleep(self.dt)
					timeout += self.dt
					if self.serialPort.inWaiting() > 0:
						resp = self.serialPort.Read(1)
						if resp[0] == self.ACK:
							logging.serial.info("Got Ack on page %i" % p)
							tryCount = 0
							break # go to next packet
						elif resp[0] == self.NAK:
							logging.serial.error("Z-Stick reported wrong frame order on page %i" % p)
							if tryCount >= self.TRY_COUNT_MAX:
								exit(3)
							self.Wait(self.TIMEOUT_WRONG_PAGE_WAIT) # wait for the chip to purge the buffer
						elif resp[0] == self.CRC:
							logging.serial.error("Z-Stick reported wrong CRC on page %i" % p)
							if tryCount >= self.TRY_COUNT_MAX:
								exit(4)
						else:
							logging.serial.error("Unknown response %i on page %i" % (resp[0], p))
							if tryCount >= self.TRY_COUNT_MAX:
								exit(6)
						tryCount += 1
						break
					if timeout >= self.TIMEOUT:
						if p == 0:
							logging.serial.info("Last page written without error. Should be done!")
							tryCount = 0
						else:
							logging.serial.error("Operation timeout on page %i" % p)
							if tryCount >= self.TRY_COUNT_MAX:
								exit(1)
							tryCount += 1
						break

		self.serialPort.Close()
		logging.program.info("Done")
		exit(0)

	def Exit(self):
		exit(1)

def Main():
	parser = optparse.OptionParser('Usage: %prog <options>', description='Z-Wave.Me Z-Stcik Updater software for Z-Wave.Me Z-Stick 4')
	parser.add_option('-d', '--device', help='path to serial port of Z-Stick', action='store', type='string', dest='device', default='/dev/ttyUSB0' if os.name == 'posix' and os.uname()[0] == 'Linux' else '/dev/tty.SLAB_USBtoUART')
	parser.add_option('-f', '--file', help='file to be written to the chip', action='store', dest='file', type='str')
	parser.add_option('-F', '--freq', help='explicitely set frequency', action='store', dest='freq', choices=('EU', 'RU', 'US'), default=None)
	option, args = parser.parse_args()

	device = SerialAPI.Port(option.device)

	ZLogging.Start('-', logLevel = logging.INFO)
                	
	if not option.file:
		logging.program.error("You must specify .hex or .ehex file to write (in Intel Hex format)")
		exit(-1)
	
	if option.freq is not None and option.file.endswith('.ehex'):
		logging.program.error("It is not possible to set frequency in .ehex files")
		exit(-2)
	
	try:
		z = Programmer(device, option.file, option.freq)
	except KeyboardInterrupt:
		logging.program.info("Exiting on keyboard interrupt")
		exit(1)

Main()
